##############################################################################
#
# Copyright (c) 2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""REST publication and publisher factories

$Id: error.py 82249 2007-12-11 00:07:36Z srichter $
"""
import zope.component
from zope.app.pagetemplate.viewpagetemplatefile import ViewPageTemplateFile
from z3c.rest import rest


class XMLErrorView(rest.RESTView):

    template = ViewPageTemplateFile('error.pt')

    def update(self):
        self.name = self.context.__class__.__name__
        self.explanation = unicode(str(self.context))

    def __call__(self):
        self.update()
        return self.template()



##############################################################################
#
# Copyright (c) 2004 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""REST Tests

$Id: test.py 82249 2007-12-11 00:07:36Z srichter $
"""
__docformat__ = "reStructuredText"
import os
import unittest
import zope.component
import zope.traversing.testing
from zope.testing import doctest, doctestunit
from zope.traversing.browser import absoluteurl
from zope.traversing.interfaces import IContainmentRoot
from zope.app.testing import functional, placelesssetup
from z3c.rest import interfaces

RESTLayer = functional.ZCMLLayer(
    os.path.join(os.path.split(__file__)[0], 'ftesting.zcml'),
    __name__, 'RESTLayer', allow_teardown=True)

def setUp(test):
    placelesssetup.setUp(test)
    zope.traversing.testing.setUp()
    
    # XXX: This really needs a REST equivalent w/o breadcrumbs.

    zope.component.provideAdapter(
        absoluteurl.AbsoluteURL,
        (None, interfaces.IRESTRequest),
        absoluteurl.IAbsoluteURL)

    zope.component.provideAdapter(
        absoluteurl.SiteAbsoluteURL,
        (IContainmentRoot, interfaces.IRESTRequest),
        absoluteurl.IAbsoluteURL,
        'absolute_url')

    zope.component.provideAdapter(
        absoluteurl.SiteAbsoluteURL,
        (IContainmentRoot, interfaces.IRESTRequest),
        absoluteurl.IAbsoluteURL)

    zope.component.provideAdapter(
        absoluteurl.AbsoluteURL,
        (None, interfaces.IRESTRequest),
        absoluteurl.IAbsoluteURL,
        'absolute_url')


def test_suite():
    client = functional.FunctionalDocFileSuite('../client.txt')
    client.layer = RESTLayer
    return unittest.TestSuite((
        client,
        doctest.DocFileSuite(
            '../rest.txt',
            globs={'pprint': doctestunit.pprint},
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
            ),
        doctest.DocFileSuite(
            '../error.txt',
            setUp=setUp, tearDown=placelesssetup.tearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
            ),
        doctest.DocFileSuite(
            '../null.txt',
            setUp=setUp, tearDown=placelesssetup.tearDown,
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
            ),
        doctest.DocFileSuite(
            '../traverser.txt',
            optionflags=doctest.NORMALIZE_WHITESPACE|doctest.ELLIPSIS,
            ),
        ))



##############################################################################
#
# Copyright (c) 2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Null resource.

$Id: null.py 82045 2007-11-30 11:40:12Z srichter $
"""
import zope.component
import zope.interface
from zope.app.http.interfaces import INullResource
from zope.location import location
from zope.traversing.browser import absoluteURL

from z3c.rest import rest

class NullResource(location.Location):
    """Object representing objects to be created by a `PUT`."""
    zope.interface.implements(INullResource)

    def __init__(self, container, name):
        self.__parent__ = self.container = container
        self.__name__ = self.name = name

    def __repr__(self):
        return '<%s %r>' %(self.__class__.__name__, self.name)


class NullPUT(rest.RESTView):
    """Put handler for null resources"""

    def PUT(self):
        nullPut = zope.component.queryMultiAdapter(
            (self.context.container, self.request), name='NullPUT')
        if nullPut is None:
            # See RFC 2616, section 9.6
            self.request.response.setStatus(501)
            return

        newObj = nullPut.NullPUT(self.context)

        # See RFC 2616, section 9.6
        self.request.response.setStatus(201)
        self.request.response.setHeader(
            'Location', absoluteURL(newObj, self.request))



##############################################################################
#
# Copyright (c) 2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test REST view for folders.

$Id: folder.py 82045 2007-11-30 11:40:12Z srichter $
"""
import lxml.etree
from z3c.rest import rest
from zope.app.folder import folder
from zope.app.pagetemplate.viewpagetemplatefile import ViewPageTemplateFile
from zope.app.publication.http import MethodNotAllowed
from zope.dublincore.interfaces import IZopeDublinCore

class FolderAPI(rest.RESTView):
    """A simple REST view for folders."""

    template = ViewPageTemplateFile("folder.pt")

    def __init__(self, context, request):
        super(FolderAPI, self).__init__(context, request)
        for param in ('noitems', 'notitle'):
            if 'HTTP_DEMO_' + param.upper() in self.request:
                self.request.parameters[param] = 1

    def GET(self):
        return self.template()

    def POST(self):
        tree = lxml.etree.parse(self.request.bodyStream)
        title = tree.find('title').text
        dc = IZopeDublinCore(self.context)
        dc.title = unicode(title)

    # For existing resources, PUT pretty much behaves like POST
    PUT = POST

    def NullPUT(self, nullResource):
        name = nullResource.name
        self.context[name] = folder.Folder()
        return self.context[name]

    def DELETE(self):
        container = self.context.__parent__
        if container is None:
            raise MethodNotAllowed(self.context, self.request)
        del container[self.context.__name__]



##############################################################################
#
# Copyright (c) 2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""REST server factories

$Id: zserver.py 82045 2007-11-30 11:40:12Z srichter $
"""
from zope.server.http.commonaccesslogger import CommonAccessLogger
from zope.server.http import wsgihttpserver
from zope.app.server.wsgi import ServerType
from zope.app.wsgi import WSGIPublisherApplication
from z3c.rest.rest import RESTPublicationRequestFactory

rest = ServerType(wsgihttpserver.WSGIHTTPServer,
                  WSGIPublisherApplication,
                  CommonAccessLogger,
                  8080, True,
                  RESTPublicationRequestFactory)



##############################################################################
#
# Copyright (c) 2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""REST publishing interfaces.

$Id: interfaces.py 82249 2007-12-11 00:07:36Z srichter $
"""
import zope.interface
import zope.schema
from zope.location.interfaces import ILocation
from zope.publisher.interfaces import http

class IRESTRequest(http.IHTTPRequest):
    """A special type of request for handling REST-based requests."""

class IRESTResponse(http.IHTTPResponse):
    """A special type of response for handling REST-based requests."""


class IRESTView(ILocation):
    """A REST view"""


class ILink(zope.interface.Interface):
    """An object representing a hyperlink."""

    href = zope.schema.TextLine(
        title=u"URL",
        description=u"The normalized URL of the link",
        required=False)

    title = zope.schema.TextLine(
        title=u'Title',
        description=u'The title of the link',
        required=False)

    def click():
        """click the link, going to the URL referenced"""


class IRESTClient(zope.interface.Interface):
    """A REST client.

    This client provides a high-level API to access RESTful Web APIs. The
    interface is designed to resemble the test browser API as much as
    practical.
    """

    connectionFactory = zope.schema.Field(
        title=u"Connection Facotry",
        description=(u'A callable that creates an `httplib`-compliant '
                     u'connection object.'),
        required=True)

    requestHeaders = zope.schema.Dict(
        title=u"Request Headers",
        description=(u"A set of headers that will be sent in every request."),
        required=True)

    url = zope.schema.URI(
        title=u"URL",
        description=(u"The URL the browser is currently showing. It is "
                     u"always a full, absolute URL."),
        required=True)

    headers = zope.schema.List(
        title=u"Response Headers",
        description=(u'A list of all headers that have been returned '
                     u'in the last request.'),
        required=True)

    contents = zope.schema.Text(
        title=u"Contents",
        description=u"The complete response body of the HTTP request.",
        required=True)

    status = zope.schema.Int(
        title=u"Status",
        description=u"The status code of the last response.",
        min=0,
        required=True)

    reason = zope.schema.TextLine(
        title=u"Reason",
        description=u"A short explanation of the status of the last response.",
        required=True)

    fullStatus = zope.schema.TextLine(
        title=u"Full Status",
        description=u"The status code and reason of the last response.",
        required=True)

    def open(url='', data=None, params=None, headers=None, method='GET'):
        """Open a URL and retrieve the result.

        The `url` argument can either be a full URL or a URL relative to the
        previous one. If no URL is specified, then the previous URL will be
        used.

        The `data` is the contents of the request body. It is used to send
        information to the server.

        The `params` describe additional query parameters that will be added
        to the request. Query string parameters are frequently used by RESTive
        APIs to provide additional return value options.

        The `headers` specify additional request headers that are specific for
        this particular request.

        The `method` specifies the HTTP method or verb to use to access the
        resource on the server. While there are only a few methods in RFC
        2616, an string is allowed, since any particular API can extend the
        set of allowed methods.
        """

    def get(url='', params=None, headers=None):
        """Make a GET request to the server.

        For argument details see ``open()``.
        """

    def put(url='', data='', params=None, headers=None):
        """Make a PUT request to the server.

        For argument details see ``open()``.
        """

    def post(url='', data='', params=None, headers=None):
        """Make a POST request to the server.

        For argument details see ``open()``.
        """

    def delete(url='', params=None, headers=None):
        """Make a DELETE request to the server.

        For argument details see ``open()``.
        """

    def setCredentials(username, password):
        """Set the credentials.

        This method adds the necessary information to authenticate the user. A
        common example is basic auth, which inserts the `Authentication`
        request header.
        """

    def goBack(count=1):
        """Go back in history by a certain amount of visisted pages.

        The ``count`` argument specifies how far to go back. It is set to 1 by
        default.
        """

    def reload():
        """Reload the current resource.

        All arguments, including the HTTP method, parameters and additional
        headers are honored.
        """

    def getLink(title=None, url=None, index=0):
        """Return an ILink of the found link.

        This method assumes that the current content type of the response body
        is XML.

        The link is found by the arguments of the method.  Only one can be
        used at a time:

          o ``title`` -- The title or a sub-string thereof of the link.

          o ``url`` -- The URL the link is going to.

        If multiple matching links are found, the `index` specifies which one
        to use.
        """

    def xpath(expr, nsmap=None, single=False):
        """Returns the result of an XPath search expression.

        This method assumes that the current content type of the response body
        is XML.

        The `expr` argument is the actual XPath expression. If the expression
        is incorrect, an unspecified error must be raised.

        The `nsmap` is a mapping from the short version to the full URL of
        each XML namespace used in the expression.

        If `single` is set to ``True``, then only one result is returned,
        instead of a list. If the XPath expression results in more than one
        result, a ``ValueError`` must be raised.
        """


class IPublisherRESTClient(IRESTClient):
    """An extension to the REST client to support test-specific features."""

    handleErrors = zope.schema.Bool(
        title=u"Handle Errors",
        description=(u"Describes whether server-side errors will be handled "
                     u"by the publisher. If set to ``False``, the error will "
                     u"progress all the way to the test, which is good for "
                     u"debugging."),
        default=True,
        required=True)



##############################################################################
#
# Copyright (c) 2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""REST Client

$Id: client.py 91201 2008-09-16 22:33:33Z srichter $
"""
import lxml
import lxml.etree
import httplib
import socket
import urllib
import urlparse
import base64
import zope.interface
from z3c.rest import interfaces

def isRelativeURL(url):
    """Determines whether the given URL is a relative path segment."""
    pieces = urlparse.urlparse(url)
    if not pieces[0] and not pieces[1]:
        return True
    return False

def absoluteURL(base, url):
    """Convertes a URL to an absolute URL given a base."""
    if not isRelativeURL(url):
        return url

    pieces = list(urlparse.urlparse(base))
    urlPieces = list(urlparse.urlparse(url))

    if not pieces[2].endswith('/'):
        pieces[2] += '/'
    pieces[2] = urlparse.urljoin(pieces[2], urlPieces[2])

    if urlPieces[4]:
        if pieces[4]:
            pieces[4] = pieces[4] + '&' + urlPieces[4]
        else:
            pieces[4] = urlPieces[4]

    return urlparse.urlunparse(pieces)


def getFullPath(pieces, params):
    """Build a full httplib request path, including a query string."""
    query = ''
    if pieces[4]:
        query = pieces[4]
    if params:
        encParams = urllib.urlencode(params)
        if query:
            query += '&' + encParams
        else:
            query = encParams
    return urlparse.urlunparse(
        ('', '', pieces[2], pieces[3], query, pieces[5]))


class XLink(object):
    """A link implementation for simple XLinks."""
    zope.interface.implements(interfaces.ILink)

    def __init__(self, client, title, url):
        self._client = client
        self.title = title
        self.url = url

    def click(self):
        """See interfaces.ILink"""
        self._client.get(self.url)

    def __repr__(self):
        return '<%s title=%r url=%r>' %(
            self.__class__.__name__, self.title, self.url)


class RESTClient(object):
    zope.interface.implements(interfaces.IRESTClient)

    connectionFactory = httplib.HTTPConnection
    sslConnectionFactory = httplib.HTTPSConnection

    def __init__(self, url=None):
        self.requestHeaders = {}
        self._reset()
        self._history = []
        self._requestData = None
        self.url = ''
        if url:
            self.open(url)

    @property
    def fullStatus(self):
        return '%i %s' %(self.status, self.reason)

    def _reset(self):
        self.headers = []
        self.contents = {}
        self.status = None
        self.reason = None

    def open(self, url='', data=None, params=None, headers=None, method='GET'):
        # Create a correct absolute URL and set it.
        self.url = absoluteURL(self.url, url)

        # Create the full set of request headers
        requestHeaders = self.requestHeaders.copy()
        if headers:
            requestHeaders.update(headers)

        # Let's now reset all response values
        self._reset()

        # Store all the request data
        self._requestData = (url, data, params, headers, method)

        # Make a connection and retrieve the result
        pieces = urlparse.urlparse(self.url)
        if pieces[0] == 'https':
            connection = self.sslConnectionFactory(pieces[1])
        else:
            connection = self.connectionFactory(pieces[1])
        try:
            connection.request(
                method, getFullPath(pieces, params), data, requestHeaders)
            response = connection.getresponse()
        except socket.error, e:
            connection.close()
            self.status, self.reason = e.args
            self._addHistory()
            raise e
        else:
            self.headers = response.getheaders()
            self.contents = response.read()
            self.status = response.status
            self.reason = response.reason
            connection.close()
            self._addHistory()

    def get(self, url='', params=None, headers=None):
        self.open(url, None, params, headers)

    def put(self, url='', data='', params=None, headers=None):
        self.open(url, data, params, headers, 'PUT')

    def post(self, url='', data='', params=None, headers=None):
        self.open(url, data, params, headers, 'POST')

    def delete(self, url='', params=None, headers=None):
        self.open(url, None, params, headers, 'DELETE')

    def setCredentials(self, username, password):
        creds = username + u':' + password
        creds = "Basic " + base64.encodestring(creds.encode('utf-8')).strip()
        self.requestHeaders['Authorization'] = creds

    def _addHistory(self):
        self._history.append((
            self.url, self.requestHeaders, self.headers, self.contents,
            self.status, self.reason, self._requestData
            ))

    def goBack(self, count=1):
        # The user really does not want to go back.
        if count == 0:
            return
        # The user wants to reach before a pre-historical state.
        if len(self._history) < count:
            raise ValueError('There is not enough history.')
        # Let's now get the entry and set the history back to that state.
        entry = self._history[-(count+1)]
        self._history = self._history[:-count]
        # Reset the state.
        (self.url, self.requestHeaders, self.headers, self.contents,
         self.status, self.reason, self._requestData) = entry

    def reload(self):
        self.open(*self._requestData)

    def getLink(self, title=None, url=None, index=0):
        nsmap = {'xlink': "http://www.w3.org/1999/xlink"}
        tree = lxml.etree.fromstring(self.contents)
        res = []
        if title is not None:
            res = tree.xpath(
                '//*[@xlink:title="%s"]' %title, namespaces=nsmap)
        elif url is not None:
            res = tree.xpath(
                '//*[@xlink:href="%s"]' %url, namespaces=nsmap)
        else:
            raise ValueError('You must specify a title or URL.')
        elem = res[index]
        url = elem.attrib.get('{%(xlink)s}href' %nsmap, '')
        return XLink(self,
                     elem.attrib.get('{%(xlink)s}title' %nsmap),
                     absoluteURL(self.url, url))

    def xpath(self, expr, nsmap=None, single=False):
        res = lxml.etree.fromstring(self.contents).xpath(expr, namespaces=nsmap)
        if not single:
            return res
        if len(res) != 1:
            raise ValueError('XPath expression returned more than one result.')
        return res[0]



##############################################################################
#
# Copyright (c) 2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""REST testing support.

$Id: testing.py 90827 2008-09-04 19:48:01Z mgedmin $
"""
import sys
import zope.interface
from z3c.rest import client, rest, interfaces
from zope.app.publication.http import HTTPPublication
from zope.app.testing.functional import HTTPCaller


class RESTCaller(HTTPCaller):
    """An HTTP caller for REST functional page tests"""

    def chooseRequestClass(self, method, path, environment):
        """Always returns HTTPRequests regardless of methods and content"""
        return rest.RESTRequest, HTTPPublication


class PublisherConnection(object):

    callerFactory = RESTCaller

    def __init__(self, server, port=None):
        self._response = None
        self.server = server
        self.port = port

    def request(self, method, path, body, headers):
        # Extract the handle_error option header
        handleErrorsKey = 'x-zope-handle-errors'
        handleErrors = headers.get(handleErrorsKey, True)
        if handleErrorsKey in headers:
            del headers[handleErrorsKey]

        # Construct the request body and call the publisher
        body = body or ''
        request = ["%s %s HTTP/1.1" % (method, path)]
        for hdr, value in headers.items():
            request.append("%s: %s" % (hdr, value))
        request_string = "\n".join(request) + "\n\n" + body
        self._response = self.callerFactory()(
            request_string, handle_errors=handleErrors)

    def getresponse(self):
        return PublisherResponse(self._response)

    def close(self):
        self._response = None


class SSLPublisherConnection(PublisherConnection):

    def __init__(self, server, port=None):
        print "Using SSL"
        PublisherConnection.__init__(self, server, port=port)


class PublisherResponse(object):
    """Adapter of Zope 3 response objects into httplib.HTTPResponse."""

    def __init__(self, response):
        self._response = response
        self.status = response.getStatus()
        self.reason = response._reason

    def getheaders(self):
        return self._response.getHeaders()

    def read(self):
        return self._response.consumeBody()


class RESTClient(client.RESTClient):
    zope.interface.implements(interfaces.IPublisherRESTClient)

    connectionFactory = PublisherConnection
    sslConnectionFactory = SSLPublisherConnection

    @apply
    def handleErrors():
        """See zope.testbrowser.interfaces.IBrowser"""
        headerKey = 'x-zope-handle-errors'

        def get(self):
            return self.requestHeaders.get(headerKey, True)

        def set(self, value):
            current_value = get(self)
            if current_value == value:
                return
            self.requestHeaders[headerKey] = value

        return property(get, set)



##############################################################################
#
# Copyright (c) 2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""REST server factories for Twisted

Note: The file is not named twisted.py because of package import problems.

$Id: twist.py 82045 2007-11-30 11:40:12Z srichter $
"""
from twisted.web2 import log, server, wsgi
from twisted.web2.channel.http import HTTPFactory
from zope.app.twisted import http
from zope.app.twisted.server import ServerType
from zope.app.wsgi import WSGIPublisherApplication
from z3c.rest.rest import RESTPublicationRequestFactory

def createRESTFactory(db):
    resource = wsgi.WSGIResource(
        WSGIPublisherApplication(db, RESTPublicationRequestFactory))
    resource = log.LogWrapperResource(resource)
    resource = http.Prebuffer(resource)

    return HTTPFactory(server.Site(resource))

rest = ServerType(createRESTFactory, 8081)



##############################################################################
#
# Copyright (c) 2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Some basic traversers for REST

$Id: traverser.py 82207 2007-12-09 04:38:37Z srichter $
"""
import zope.interface
from z3c.rest import interfaces, null
from z3c.traverser import traverser
from z3c.traverser.interfaces import ITraverserPlugin
from zope.app.container.interfaces import IItemContainer
from zope.publisher.interfaces import NotFound


class RESTPluggableTraverser(traverser.BasePluggableTraverser):
    """A simple REST-compliant pluggable traverser."""


class ContainerItemTraverserPlugin(object):
    """A traverser that knows how to look up objects by name in a container."""
    zope.interface.implements(ITraverserPlugin)
    zope.component.adapts(IItemContainer, interfaces.IRESTRequest)

    def __init__(self, container, request):
        self.context = container
        self.request = request

    def publishTraverse(self, request, name):
        """See zope.publisher.interfaces.IPublishTraverse"""
        try:
            return self.context[name]
        except KeyError:
            return self.nullResource(request, name)

    def nullResource(self, request, name):
        # we traversed to something that doesn't exist.

        # The name must be the last name in the path, so the traversal
        # name stack better be empty:
        if request.getTraversalStack():
            raise NotFound(self.context, name, request)

        # This should only happen for a PUT:
        if request.method != 'PUT':
            raise NotFound(self.context, name, request)

        return null.NullResource(self.context, name)



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Setup

$Id: setup.py 91201 2008-09-16 22:33:33Z srichter $
"""
import os
from setuptools import setup, find_packages

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

chapters = '\n'.join(
    [read('src', 'z3c', 'rest', name)
    for name in ('README.txt',
                 'client.txt',
                 'null.txt',
                 'traverser.txt',
                 'rest.txt')])

setup (
    name='z3c.rest',
    version='0.2.5',
    author = "Stephan Richter and the Zope Community",
    author_email = "zope3-dev@zope.org",
    description = "A REST Framework for Zope 3 Applications",
    long_description=(
        read('README.txt')
        + '\n\n' +
        'Detailed Documentation\n'
        '**********************\n'
        + '\n' + chapters
        + '\n\n' +
        read('CHANGES.txt')
        ),
    license = "ZPL 2.1",
    keywords = "zope3 form widget",
    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Framework :: Zope3'],
    url = 'http://cheeseshop.python.org/pypi/z3c.rest',
    packages = find_packages('src'),
    include_package_data = True,
    package_dir = {'':'src'},
    namespace_packages = ['z3c'],
    extras_require = dict(
        app = ['zope.app.appsetup',
               'zope.app.authentication',
               'zope.app.component',
               'zope.app.container',
               'zope.app.error',
               'zope.app.form',
               'zope.app.publisher',
               'zope.app.publication',
               'zope.app.security',
               'zope.app.securitypolicy',
               'zope.app.twisted',
               'zope.app.wsgi',
               'zope.app.zcmlfiles',
               'zope.contentprovider',
               ],
        test = ['z3c.coverage',
                'z3c.etestbrowser',
                'zope.app.testing'],
        ),
    install_requires = [
        'lxml>=2.0.0', # Changes in API, since 1.3.6.
        'setuptools',
        'z3c.traverser',
        'zope.publisher',
        ],
    zip_safe = False,
    )



##############################################################################
#
# Copyright (c) 2007 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""REST publication and publisher factories

$Id: rest.py 82249 2007-12-11 00:07:36Z srichter $
"""
import cgi
import types
import zope.interface
from z3c.rest import interfaces
from zope.pagetemplate.pagetemplatefile import PageTemplateFile
from zope.app.publication.interfaces import IPublicationRequestFactory
from zope.app.publication.http import HTTPPublication
from zope.publisher.http import HTTPRequest, HTTPResponse
from zope.publisher.interfaces import Redirect


class RESTRequest(HTTPRequest):
    zope.interface.implements(interfaces.IRESTRequest)

    __slots__ = (
        'parameters', # Parameters sent via the query string.
        )

    def __init__(self, body_instream, environ, response=None):
        self.parameters = {}
        super(RESTRequest, self).__init__(body_instream, environ, response)

    def processInputs(self):
        'See IPublisherRequest'
        if 'QUERY_STRING' not in self._environ:
            return
        # Parse the query string into our parameters dictionary.
        self.parameters = cgi.parse_qs(
            self._environ['QUERY_STRING'], keep_blank_values=1)
        # Since the parameter value is always a list (sigh), let's at least
        # detect single values and store them.
        for name, value in self.parameters.items():
            if len(value) == 1:
                self.parameters[name] = value[0]

    def keys(self):
        'See Interface.Common.Mapping.IEnumerableMapping'
        d = {}
        d.update(self._environ)
        d.update(self.parameters)
        return d.keys()

    def get(self, key, default=None):
        'See Interface.Common.Mapping.IReadMapping'
        marker = object()
        result = self.parameters.get(key, marker)
        if result is not marker:
            return result

        return super(RESTRequest, self).get(key, default)

    def _createResponse(self):
        return RESTResponse()


class RESTResponse(HTTPResponse):
    zope.interface.implements(interfaces.IRESTResponse)

    errorTemplate = PageTemplateFile("baseerror.pt")

    def handleException(self, exc_info):
        errorClass, error = exc_info[:2]
        if isinstance(errorClass, (types.ClassType, type)):
            if issubclass(errorClass, Redirect):
                self.redirect(error.getLocation())
                return
            title = tname = errorClass.__name__
        else:
            title = tname = unicode(errorClass)

        # Throwing non-protocol-specific exceptions is a good way
        # for apps to control the status code.
        self.setStatus(tname)

        self.setHeader("Content-Type", "text/xml")
        self.setResult(
            self.errorTemplate(name=title, explanation=str(error)))


class RESTView(object):
    zope.interface.implements(interfaces.IRESTView)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    @apply
    def __parent__():
        def get(self):
            return getattr(self, '_parent', self.context)
        def set(self, parent):
            self._parent = parent
        return property(get, set)


class RESTPublicationRequestFactory(object):
    zope.interface.implements(IPublicationRequestFactory)

    def __init__(self, db):
        """See zope.app.publication.interfaces.IPublicationRequestFactory"""
        self.publication = HTTPPublication(db)

    def __call__(self, input_stream, env, output_stream=None):
        """See zope.app.publication.interfaces.IPublicationRequestFactory"""
        request = RESTRequest(input_stream, env)
        request.setPublication(self.publication)
        return request
